/*
Title:		rc_mosaic-cgi.c
Author:		Shawn A. Clifford
Date:		16-Sep-1994
Purpose:	This Common Gateway Interface (CGI) program will parse
		the command and URL to be issued to the slave Mosaic.
		The command and URL are passed as "extra path information",
		and should be delimited by single quotes.
		
   Example:
		
	http://alpha/cgi-bin/rc_mosaic-cgi/'newwin'/'http://sun-robot/~sac/'


		In this example, the command is 'newwin' and the URL is
		'http://sun-robot/~sac/'.  'newwin' and 'goto' are the two
		supported commands under NCSA Mosaic.  See RC_MOSAIC.C for
		a list of commands supported by Bjorn Nilsson's VMS port.
		
		This program communicates with the  RC_MOSAIC  server to get
		the information necessary to trigger the slave Mosaic to load
		the specified URL.  On Unix machines this means that RC_MOSAIC
		will give us the pid of the slave Mosaic so that we can create
		the /temp/Mosaic.<pid> control file, and issue a SIGUSR1 signal
		to the slave Mosaic, causing it to interpret the file.  On VMS
		machines we will get the mailbox to be used to communicate the
		command and URL to the slave Mosaic.
*/


#include "rc_mosaic.h"			/* Common defines for RC_MOSAIC */

#include <stdio.h>			/* printf(), sprintf(), etc. */
#include <stdlib.h>			/* most of the C routines */
#include <string.h>			/* For strtok() */


#ifdef VMS
#	include <errno.h>
#	include <types.h>
#	include <socket.h>
#	include <in.h>
#	include <netdb.h>		/* change hostent to comply w/ BSD 4.3*/
#	include <inet.h>
#	include <ucx$inetdef.h>		/* INET symbol definitions */
#	include <unixio.h>		/* close() */
#else
#	include <errno.h>
#	include <sys/types.h>
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <netdb.h>
#	include <arpa/inet.h>
#	include <sys/uio.h>
#endif




/*
Define this if you are using the Ohio State DECthreaded Web Server on
a VMS machine.  You may also CC/DEFINE=VMS_OSU_SERVER
*/

/* #define VMS_OSU_SERVER	1	/* VMS Ohio State Web Server */


#ifdef VMS_OSU_SERVER
#	include	"cgilib.h"	/* For CGI environment on VMS OSU WWW server */
#	define	printf	cgi_printf
#endif

/* This will make your DECthreads CGI code CERN compliant. */

#ifdef NOCGILIB
#	define cgi_init(a,b)  1
#	define cgi_info(a) getenv(strcpy(&cgi_info_buf[4], a)-4)
	static char cgi_info_buf[64] = { 'W', 'W', 'W', '_' };
#	define cgi_printf printf
#endif


#ifndef VMS
#	include <signal.h>
	static int	pid;	/* The pid which we will signal with SIGUSR1 */
#endif



/* Function prototypes */

void cleanup(int, int);



/* Global variables */

static char	delims[]  = "'";	/* Token delimeter is "'"*/



					
#ifdef VMS
int main(int argc, char **argv)
#else
int main()			/* Sun CC compiler doesn't like non-K&R C ? */
#endif
{
char path_info[255];		/* The extra path info passed to us */
char *rc_cmd;			/* The command to be sent to the Mosaic */
char *rc_URL;			/* The URL of the command sent to Mosaic */
char *junk;			/* Junk between delimeters */

int     sock_1;               	/* socket */
char  	tmp[80];		/* The pid or mailbox from RC_MOSAIC  */
static 	struct  sockaddr_in sock2_name;	/* Address struct for socket2.*/
struct  hostent         hostentstruct;	/* Storage for hostent data.  */
struct  hostent         *hostentptr;	/* Pointer to hostent data.   */
static  char            hostname[256];	/* Name of local host.        */
int	flag;			/* Receive flag */
int	retval;			/* Helpful for debugging */
int	shut = FALSE;		/* Flag to cleanup */

FILE	*fp;			/* Mr. File Pointer */


#ifdef VMS_OSU_SERVER
   	/* Get the passed command and URL */
	cgi_init ( argc, argv );
#endif

	/* Initialize buffers */   
   	memset(path_info, 0, sizeof(path_info));
	rc_cmd = rc_URL = junk = path_info;

   
#ifdef VMS_OSU_SERVER
   	if (cgi_info("PATH_INFO") != NULL) {
   		strncpy(path_info, cgi_info("PATH_INFO"),sizeof(path_info));
#else
   	if (getenv("PATH_INFO") != NULL) {
   		strncpy(path_info, getenv("PATH_INFO"),sizeof(path_info));
#endif
   		junk = strtok(path_info, delims);
		rc_cmd = strtok(NULL, delims);		/* The command */
   		junk = strtok(NULL, delims);
   		rc_URL = strtok(NULL, delims);		/* The URL */
	}

	if (rc_cmd == NULL) rc_cmd = path_info;	/* These should *NOT* be */
	if (rc_URL == NULL) rc_URL = path_info; /* NULL if the URL extra */
						/* info was correctly    */
						/* quoted. /'newwin'/'url' */



	/*** Now we want to talk to the RC_MOSAIC to get the pid/mailbox ***/

	/*
	 * Open socket 1: AF_INET, SOCK_STREAM. 
	 */
        if ((sock_1 = socket (AF_INET, SOCK_STREAM, 0)) == -1) 
		{
                perror( "socket");
		exit(1);
		}

	/*
	 * Get the host local name. 
	 */
        retval = gethostname(hostname,sizeof hostname);
	if (retval)
		{
		perror ("gethostname");
		cleanup (1, sock_1);
		}

	/*
	 * Get pointer to network data structure for socket 2 (remote host). 
	 */
        if ((hostentptr = gethostbyname (hostname)) == NULL) 
		{
                perror( "gethostbyname");
		cleanup(shut, sock_1);
		}

	/*
	 * Copy hostent data to safe storage. 
	 */
        hostentstruct = *hostentptr;

	/*
	 * Fill in the name & address structure for socket 2. 
	 */
        sock2_name.sin_family = hostentstruct.h_addrtype;
	sock2_name.sin_port = htons(RC_MOSAIC_PORT);
        sock2_name.sin_addr = * ((struct in_addr *) hostentstruct.h_addr);

	/*
	 * Connect socket 1 to sock2_name.
	 */
	retval = connect(sock_1, (struct sockaddr *) &sock2_name,
			sizeof(sock2_name));
	if (retval)
		{
		perror("connect");
		cleanup(shut, sock_1);
		}

	/*
	 * Receive packet
	 */
	flag = 0;	/* maybe 0 or MSG_OOB or MSG_PEEK */
	memset (tmp, 0, sizeof(tmp));
	retval = recv(sock_1, tmp, sizeof (tmp), flag);
	if (retval == -1)
		{
		perror ("receive");
		shut = TRUE;
		}



   
/* ------  Spit out the HTML header ------- */   
   
	printf("Content-type: text/html\n\n");
	
	printf("<HTML>\n\n<HEAD>\n<TITLE>RC_MOSAIC-CGI</TITLE>\n</HEAD>\n\n");
   
	printf("<BODY>\n\n<P>");



/* Do it. */
	   
/* I'm putting this here so that the printf will come back to Mosaic. */

	if ((fp = fopen(tmp, "w")) == NULL) {
	   printf("Error opening %s.  RC_MOSAIC must be running.", tmp);
	   exit(1);
	}

#ifndef VMS

	/* This is Un*x Mosaic.  Open the temp file and write it out. */
	   
	fprintf(fp, "%s\n%s", rc_cmd, rc_URL);
	fflush(fp);
	fclose(fp);

        /* Send the SIGUSR1 signal to the RC_DISPLAY process so that it */
	/* will then open/read/process the /tmp/Mosaic.<pid> file.      */

	sscanf(tmp, "/tmp/Mosaic.%d", &pid);	/* Extract the pid */
	kill(pid, SIGUSR1);
	   
#else	/* This is VMS Mosaic.  Write to the mailbox this way. */
	   
	fprintf(fp, "%s|%s", rc_cmd, rc_URL);
	fflush(fp);
	fclose(fp);
	   
#endif


   	printf("Sent '%s|%s' to RC_DISPLAY Mosaic via %s\n",rc_cmd,rc_URL,tmp);

   	printf("\n<BR>\n\n");
	printf("Use your browser's 'Back' button to return to the previous \
page.");
   	printf("\n<BR><BR><HR><BR>\n\n");
   	printf("<ADDRESS>\nRC_MOSAIC CGI / UF / \n");

        printf("<A href=\"http://sun-robot.nuceng.ufl.edu/~sac/\">\
Shawn A. Clifford &lt;sac@ufl.edu&gt;</A>");

        printf("<BR>\n07Mar1995\n</ADDRESS>\n");
   	printf("</P>\n\n</BODY>\n\n</HTML>");

/* ------- end of HTML -------- */



   /*
    * Call cleanup to shutdown and close socket and exit.
    */
   cleanup(shut, sock_1);

}	/* main() */




void cleanup(int shut, int socket)
{
	int	retval;

	/*
	 * Shutdown socket completely -- only if it was connected 
	 */
	if (shut) {
		retval = shutdown(socket,2);
		if (retval == -1)
			perror ("shutdown");
	}


	/*
	 * Close socket.
	 */
        retval = close (socket);
	if (retval)
		perror ("close");

	exit(1);
}
